# Copyright (c) IBM Corporation
# SPDX-License-Identifier: MIT

import logging

from ado_actuators.vllm_performance.k8s.manage_components import (
    ComponentsManager,
)
from ado_actuators.vllm_performance.k8s.yaml_support.build_components import (
    ComponentsYaml,
    VLLMDtype,
)

logger = logging.getLogger(__name__)


def create_test_environment(
    k8s_name: str,
    model: str,
    pvc_name: str,
    in_cluster: bool = True,
    verify_ssl: bool = False,
    image: str = "vllm/vllm-openai:v0.6.3",
    image_secret: str = "",
    deployment_template: None | str = None,
    service_template: None | str = None,
    n_gpus: int = 1,
    gpu_type: str = "NVIDIA-A100-80GB-PCIe",
    node_selector: dict[str, str] = {},
    n_cpus: int = 8,
    memory: str = "128Gi",
    max_batch_tokens: int = 16384,
    gpu_memory_utilization: float = 0.9,
    dtype: VLLMDtype = VLLMDtype.AUTO,
    cpu_offload: int = 0,
    max_num_seq: int = 256,
    hf_token: str | None = None,
    reuse_service: bool = True,
    reuse_deployment: bool = True,
    namespace: str = "vllm-testing",
    check_interval: int = 5,
    timeout: int = 1200,
) -> None:
    """
    Create test deployment
    :param k8s_name: unique K8s name
    :param model: LLM model name
    :param namespace: namespace to use for deployment
    :param pvc_name: name of the pvc to be used
    :param in_cluster: flag - running in cluster
    :param verify_ssl:  flag - verify ssl
    :param image: image to use in deployment
    :param image_secret: name of the image pull secret
    :param deployment_template: deployment template
    :param service_template: service template
    :param n_gpus: number of GPUs
    :param gpu_type: type of the GPU to use
    :param node_selector: optional node selector
    :param n_cpus: number of CPUs
    :param memory: pod memory
    :param max_batch_tokens: Vllm parameter - maximum number of batched tokens per iteration
    :param gpu_memory_utilization: VLLM parameter - GPU memory utilization
    :param dtype: VLLM parameter - data type for model weights and activations
    :param cpu_offload: VLLM parameter - the space in GiB to offload to CPU, per GPU
    :param max_num_seq: VLLM parameter - Maximum number of sequences per iteration.
    :param hf_token: huggingface token
    :param reuse_service: flag to reuse deployment
    :param reuse_deployment: flag to reuse deployment
    :param check_interval: wait interval in seconds
    :param timeout: timeout in seconds
    :return:
    """
    logger.info(f"Creating environment in ns {namespace} with the parameters: ")
    logger.info(
        f"model {model}, in_cluster {in_cluster}, verify_ssl {verify_ssl}, image {image}"
    )
    logger.info(
        f"image_secret {image_secret}, deployment_template {deployment_template}, "
        f"service_template {service_template}, pvc_name {pvc_name}"
    )
    logger.info(f"n_gpus {n_gpus}, gpu_type {gpu_type}, n_cpus {n_cpus}")
    logger.info(f"node selector {node_selector}")
    logger.info(
        f"memory {memory}, max_batch_tokens {max_batch_tokens}, gpu_memory_utilization {gpu_memory_utilization}"
    )
    logger.info(f"dtype {dtype}, cpu_offload {cpu_offload}, max_num_seq {max_num_seq}")
    logger.info(f"reuse_service {reuse_service}, reuse_deployment {reuse_deployment}")

    # manager
    c_manager = ComponentsManager(
        namespace=namespace,
        in_cluster=in_cluster,
        verify_ssl=verify_ssl,
    )
    logger.debug("component manager created")

    # deployment
    c_manager.create_deployment(
        k8s_name=k8s_name,
        model=model,
        gpu_type=gpu_type,
        node_selector=node_selector,
        image=image,
        image_secret=image_secret,
        n_gpus=n_gpus,
        n_cpus=n_cpus,
        memory=memory,
        max_batch_tokens=max_batch_tokens,
        gpu_memory_utilization=gpu_memory_utilization,
        dtype=dtype,
        cpu_offload=cpu_offload,
        max_num_seq=max_num_seq,
        template=deployment_template,
        claim_name=pvc_name,
        hf_token=hf_token,
        reuse=reuse_deployment,
    )
    logger.debug("deployment created")
    c_manager.wait_deployment_ready(
        k8s_name=k8s_name,
        check_interval=check_interval,
        timeout=timeout,
    )
    logger.info("deployment ready")
    # service
    c_manager.create_service(
        k8s_name=k8s_name, template=service_template, reuse=reuse_service
    )
    logger.info("service created")


if __name__ == "__main__":
    t_model = "meta-llama/Llama-3.1-8B-Instruct"
    create_test_environment(
        k8s_name=ComponentsYaml.get_k8s_name(model=t_model),
        in_cluster=False,
        verify_ssl=False,
        model=t_model,
        pvc_name="vllm-support",
        hf_token="token",
        image="quay.io/dataprep1/data-prep-kit/vllm_image:0.1",
    )



# Copyright (c) IBM Corporation
# SPDX-License-Identifier: MIT

import json
import os
from typing import Any


class VLLMBenchmarkResultReadError(Exception):
    """Raised if there was an issue reading benchmark results"""


def get_results(f_name: str = "random.json") -> dict[str, Any]:
    """
    Get benchmark results
    :param f_name: file containing results
    :return: results dictionary
    """
    try:
        with open(f_name) as f:
            results = json.load(f)
        os.remove(f_name)
    except Exception as e:
        raise VLLMBenchmarkResultReadError(
            f"Failed to read benchmark result due to {e}"
        )
    del results["date"]
    del results["endpoint_type"]
    del results["tokenizer_id"]
    del results["label"]
    # The CLI invocation does not return the same dict as the script invocation
    # so the following lines had to be commented
    # ---- uncomment if directly invoking script
    # del results["backend"]
    # del results["best_of"]
    # del results["request_goodput:"]
    # del results["input_lens"]
    # del results["output_lens"]
    # del results["ttfts"]
    # del results["itls"]
    # del results["generated_texts"]
    # del results["errors"]
    return results


if __name__ == "__main__":
    get_results()



# Copyright (c) IBM Corporation
# SPDX-License-Identifier: MIT

import json
import logging
import math
import time
import uuid

from ado_actuators.vllm_performance.k8s import (
    K8sConnectionError,
)
from ado_actuators.vllm_performance.k8s.yaml_support.build_components import (
    ComponentsYaml,
    VLLMDtype,
)
from kubernetes import client, config
from kubernetes.client import ApiException, V1Deployment

logger = logging.getLogger(__name__)

# These are the most common reasons for a container failure.
container_waiting_error_reasons = [
    "CrashLoopBackOff",
    "ImagePullBackOff",
    "ErrImagePull",
]


class ComponentsManager:
    """
    This class manages K8s operations
    """

    def __init__(
        self,
        namespace: str = "discovery-dev",
        in_cluster: bool = True,
        verify_ssl: bool = True,
        init_pvc: bool = False,
        pvc_name: None | str = None,
        pvc_template: None | str = None,
    ):
        """
        set up for configuration usage
        :param namespace: cluster namespace to use
        :param in_cluster: flag defining whether we are running in cluster
        :param verify_ssl: flag to verify SSL (self-signed certificate)
        :param init_pvc: flag to decide whether to initialize the PVC for the experiment
        :param pvc_name: the name of the pvc to be created
        :param pvc_template: the name of the template file for creating PVCs
        """
        try:
            if in_cluster:
                config.load_incluster_config()
            else:
                config.load_kube_config()
            if not verify_ssl:
                configuration = client.configuration.Configuration.get_default_copy()
                configuration.verify_ssl = False
                client.Configuration.set_default(configuration)
            self.kube_client_V1 = client.CoreV1Api()
            self.kube_client = client.AppsV1Api()
            # this is just to make sure we are authenticated to the cluster
            # and fail immediately otherwise.
            self.kube_client_V1.list_namespaced_pod(namespace=namespace)
        except ApiException:
            error_message = "Error connecting to the Kubernetes cluster.\n"
            if in_cluster:
                error_message += "Make sure the service account used for the Ray cluster has sufficient permissions."
            else:
                error_message += "Make sure you are authenticated with the Kubernetes cluster or that you are using a valid kubeconfig."
            logger.critical(error_message)
            raise K8sConnectionError
        except Exception as e:
            logger.critical(f"Exception connecting to kubernetes {e}")
            raise

        self.namespace = namespace

        # We do this only once when creating a ComponentsManager in the EnvironmentManager because
        # we want the PVC to be shared by all deployments we are testing with the same operation.
        if init_pvc:
            if pvc_name is None:
                # Make sure the PVC we create has a unique name
                self.pvc_name = f"vllm-support-{uuid.uuid4().hex!s}"
                self.create_pvc(pvc_name=self.pvc_name, template=pvc_template)
                self.pvc_created = True
                logger.debug(f"Created pvc {pvc_name} in namespace {namespace}")
            else:
                if not self.check_pvc_exists(pvc_name=pvc_name):
                    error_message = (
                        f"The PVC {pvc_name} does not exist in namespace {namespace}."
                    )
                    logger.error(error_message)
                    raise ValueError(error_message)
                self.pvc_name = pvc_name
                self.pvc_created = False
                logger.debug(f"Reusing pvc {pvc_name} from namespace {namespace}")

    @classmethod
    def verify_k8s_auth(cls, namespace: str) -> bool:
        try:
            version_api = client.VersionApi()
            # Fetch cluster version to verify we are connected to the cluster
            version_info = version_api.get_code()
            api = client.CoreV1Api()
            api.list_namespaced_pod(namespace)
            print("Connected to Kubernetes cluster!")
            print(f"Cluster Version: {version_info.git_version}")
            print(f"Platform: {version_info.platform}")

            return True
        except ApiException:
            print("api exception")
            return False
        except Exception:
            print("Failed connecting to the Kubernetes cluster")
            raise K8sConnectionError

    def check_pvc_exists(self, pvc_name: str) -> bool:
        """
        Check if PVC exists
        :param pvc_name: pvc name
        :return: boolean
        """
        try:
            pvcs = self.kube_client_V1.list_namespaced_persistent_volume_claim(
                namespace=self.namespace
            )
        except ApiException as e:
            logger.error(f"error getting pvc list {e}")
            return False
        return any(pvc.metadata.name == pvc_name for pvc in pvcs.items)

    def delete_pvc(self) -> None:
        """
        Delete service for model
        :param pvc_name: pvc name
        :return: boolean
        """
        try:
            self.kube_client_V1.delete_namespaced_persistent_volume_claim(
                namespace=self.namespace, name=self.pvc_name
            )
        except ApiException as e:
            logger.error(f"error deleting pvc {e}")

        logger.debug(f"Deleted pvc {self.pvc_name} from namespace {self.namespace}")

    def create_pvc(self, pvc_name: str, template: None | str = None) -> None:
        """
        create service for model
        :param pvc_name: pvc name
        :param template: yaml template name
        :return:
        """
        try:
            self.kube_client_V1.create_namespaced_persistent_volume_claim(
                namespace=self.namespace,
                body=ComponentsYaml.pvc_yaml(pvc_name=pvc_name, template=template),
            )
        except ApiException as e:
            logger.error(f"error creating pvc  {e}")
            raise

    def check_service_exists(self, k8s_name: str) -> bool:
        """
        Check if service for model exists
        :param k8s_name: kubernetes name
        :return: boolean
        """
        try:
            svcs = self.kube_client_V1.list_namespaced_service(namespace=self.namespace)
        except ApiException as e:
            logger.error(f"error getting service list {e}")
            return False
        return any(svc.metadata.name == k8s_name for svc in svcs.items)

    def delete_service(self, k8s_name: str) -> None:
        """
        Delete service for model
        :param k8s_name: kubernetes name
        :return: boolean
        """
        self.kube_client_V1.delete_namespaced_service(
            namespace=self.namespace,
            name=k8s_name,
        )

    def create_service(
        self, k8s_name: str, template: str | None = None, reuse: bool = False
    ) -> None:
        """
        create service for model
        :param k8s_name: kubernetes name
        :param template service yaml template
        :param reuse: reuse if exists
        :return:
        """
        # try to reuse existing one if exists
        exists = self.check_service_exists(k8s_name=k8s_name)
        if exists and reuse:
            return
        if exists and not reuse:
            # delete it first
            try:
                self.delete_service(k8s_name=k8s_name)
            except ApiException as e:
                logger.error(f"Error deleting service {e}")
            # make sure that deletion is completed
            deleting = True
            for _ in range(150):
                deleting = self.check_service_exists(k8s_name=k8s_name)
                if not deleting:
                    break
                time.sleep(1)
            if deleting:
                logger.error("Did not complete Service deletion")
        # create service
        try:
            self.kube_client_V1.create_namespaced_service(
                namespace=self.namespace,
                body=ComponentsYaml.service_yaml(k8s_name=k8s_name, template=template),
            )
        except ApiException as e:
            logger.error(f"error creating service  {e}")
            raise

    def check_deployment_exist(self, k8s_name: str) -> bool:
        """
        Check if deployment for model exists
        :param k8s_name: kubernetes name
        :return: boolean
        """
        try:
            deployments = self.kube_client.list_namespaced_deployment(
                namespace=self.namespace
            )
        except ApiException as e:
            logger.error(f"error getting deployment list {e}")
            return False
        for deployment in deployments.items:
            if deployment.metadata.name == k8s_name:
                return True
        return False

    def delete_deployment(self, k8s_name: str) -> None:
        """
        Delete service for model
        :param k8s_name: kubernetes name
        :return: boolean
        """
        self.kube_client.delete_namespaced_deployment(
            namespace=self.namespace,
            name=k8s_name,
            body=client.V1DeleteOptions(
                propagation_policy="Foreground", grace_period_seconds=5
            ),
        )

    def create_deployment(
        self,
        k8s_name: str,
        model: str,
        gpu_type: str = "NVIDIA-A100-80GB-PCIe",
        node_selector: dict[str, str] = {},
        image: str = "vllm/vllm-openai:v0.6.3",
        image_secret: str = "",
        n_gpus: int = 1,
        n_cpus: int = 8,
        memory: str = "128Gi",
        max_batch_tokens: int = 16384,
        gpu_memory_utilization: float = 0.9,
        dtype: VLLMDtype = VLLMDtype.AUTO,
        cpu_offload: int = 0,
        max_num_seq: int = 256,
        template: str | None = None,
        claim_name: str | None = None,
        hf_token: str | None = None,
        reuse: bool = False,
    ) -> None:
        """
        create deployment for model
        :param k8s_name: kubernetes name
        :param model: LLM model name
        :param gpu_type: gpu type, for example NVIDIA-A100-80GB-PCIe, Tesla-V100-PCIE-16GB, etc.
        :param node_selector: optional node selector
        :param image: image name to use
        :param image_secret: name of the image pull secret
        :param n_gpus: number of GPUs to use in VLLM
        :param n_cpus: number of CPUs for VLLM pod
        :param memory: memory for VLLM pod
        :param max_batch_tokens: Vllm parameter - maximum number of batched tokens per iteration
        :param gpu_memory_utilization: VLLM parameter - GPU memory utilization
        :param dtype: VLLM parameter - data type for model weights and activations
        :param cpu_offload: VLLM parameter - the space in GiB to offload to CPU, per GPU
        :param max_num_seq: VLLM parameter - Maximum number of sequences per iteration.
        :param template: template for deployment yaml
        :param claim_name: PVC name
        :param hf_token: huggingface token
        :param reuse: reuse if exists
        :return:
        """
        # try to reuse existing one if exists
        exists = self.check_deployment_exist(k8s_name=k8s_name)
        if exists and reuse:
            return
        if exists and not reuse:
            # delete it first
            try:
                self.delete_deployment(k8s_name=k8s_name)
            except ApiException as e:
                logger.error(f"Error deleting deployment {e}")
            # make sure that deletion is completed
            deleting = True
            for _ in range(150):
                deleting = self.check_deployment_exist(k8s_name=k8s_name)
                if not deleting:
                    break
                time.sleep(1)
            if deleting:
                logger.error("Did not complete deployment deletion")
                raise

        # create deployment
        deployment_yaml = ComponentsYaml.deployment_yaml(
            k8s_name=k8s_name,
            model=model,
            gpu_type=gpu_type,
            node_selector=node_selector,
            image=image,
            image_secret=image_secret,
            n_gpus=n_gpus,
            n_cpus=n_cpus,
            memory=memory,
            max_batch_tokens=max_batch_tokens,
            gpu_memory_utilization=gpu_memory_utilization,
            dtype=dtype,
            cpu_offload=cpu_offload,
            max_num_seq=max_num_seq,
            template=template,
            claim_name=claim_name,
            hf_token=hf_token,
        )
        logger.debug(json.dumps(deployment_yaml, indent=2))

        try:
            self.kube_client.create_namespaced_deployment(
                namespace=self.namespace,
                body=deployment_yaml,
            )
        except ApiException as e:
            logger.error(f"error creating deployment  {e}")
            raise

    def _is_pod_failed(self, deployment: V1Deployment) -> bool:
        label_selector = ",".join(
            [f"{k}={v}" for k, v in deployment.spec.selector.match_labels.items()]
        )
        pods = self.kube_client_V1.list_namespaced_pod(
            self.namespace, label_selector=label_selector
        )
        # There's really only one pod in our deployments
        for pod in pods.items:
            pod_name = pod.metadata.name
            pod_phase = pod.status.phase

            # Check if pod phase is Failed
            if pod_phase == "Failed":
                logger.warning(f"Pod {pod_name} is Failed")
                return True

            # Check container statuses for errors
            if pod.status.container_statuses:
                for cs in pod.status.container_statuses:
                    if (
                        cs.state.waiting
                        and cs.state.waiting.reason in container_waiting_error_reasons
                    ):
                        logger.warning(
                            f"Container {cs.name} in pod {pod_name} is in error ({cs.state.waiting.reason})"
                        )
                        return True
        return False

    def _deployment_ready(self, k8s_name: str) -> bool:
        """
        Check whether deployment pod ready
        :param k8s_name: kubernetes name
        :return: boolean
        """
        try:
            deployment = self.kube_client.read_namespaced_deployment(
                namespace=self.namespace,
                name=k8s_name,
            )
        except ApiException as e:
            logger.error(f"error getting deployment  {e}")
            return False
        if self._is_pod_failed(deployment=deployment):
            return False
        if deployment.status.available_replicas is None:
            return False
        return deployment.status.available_replicas == 1

    def wait_deployment_ready(
        self, k8s_name: str, check_interval: int = 5, timeout: int = 1200
    ) -> None:
        """
        Wait for deployment to become ready
        :param k8s_name: kubernetes name
        :param check_interval: wait interval
        :param timeout: timeout
        :return: None
        """
        n_checks = math.ceil(timeout / check_interval)
        for _ in range(n_checks):
            time.sleep(check_interval)
            if self._deployment_ready(k8s_name=k8s_name):
                return
        logger.error("Timed out waiting for deployment to get ready")
        raise Exception("Timed out waiting for deployment to get ready")


if __name__ == "__main__":
    # model
    t_model = "meta-llama/Llama-3.1-8B-Instruct"
    t_k8s_name = ComponentsYaml.get_k8s_name(model=t_model)
    # manager
    c_manager = ComponentsManager(in_cluster=False, verify_ssl=False)
    # pvc
    c_manager.check_pvc_exists(pvc_name="vllm-support")
    # service
    c_manager.create_service(k8s_name=t_k8s_name, reuse=False)
    # deployment
    c_manager.create_deployment(
        k8s_name=t_k8s_name,
        model="meta-llama/Llama-3.1-8B-Instruct",
        claim_name="vllm-support",
        hf_token="token",
        image="quay.io/dataprep1/data-prep-kit/vllm_image:0.1",
    )
    c_manager.wait_deployment_ready(k8s_name=t_k8s_name)
    logger.info("environment is created")



# Copyright (c) IBM Corporation
# SPDX-License-Identifier: MIT

import asyncio
import logging

import ray

from orchestrator.modules.operators.console_output import RichConsoleSpinnerMessage

logger = logging.getLogger(__name__)


class DeploymentWaiter:
    def __init__(self, k8s_name: str):
        self.k8s_name = k8s_name
        self.model_downloaded_event = asyncio.Event()


class DeploymentConflictManager:
    def __init__(self):
        self.deployments_to_wait_for: dict[str, DeploymentWaiter] = {}
        self.model_already_downloaded: set[str] = set()

    def maybe_add_deployment(self, model: str, k8s_name: str) -> bool:
        if (
            model in self.model_already_downloaded
            or model in self.deployments_to_wait_for
        ):
            return False

        self.deployments_to_wait_for[model] = DeploymentWaiter(k8s_name=k8s_name)
        return True

    async def wait(self, request_id: str, k8s_name: str, model: str) -> None:
        waiter = self.deployments_to_wait_for.get(model, None)
        # making sure a deployment does not wait for itself to be READY
        if waiter is not None and waiter.k8s_name != k8s_name:
            console = ray.get_actor(name="RichConsoleQueue")
            while True:
                console.put.remote(
                    message=RichConsoleSpinnerMessage(
                        id=request_id,
                        label=f"({request_id}) Waiting on deployment ({waiter.k8s_name}) to download the model required for this experiment",
                        state="start",
                    )
                )
                await waiter.model_downloaded_event.wait()
                # If after we got awoken the model is not among the downloaded models, it means that
                # something has gone wrong, such as the deployment we were waiting for has failed.
                # If am the first to wake up let me add myself as the deployment to be waited for and stop waiting.
                if (
                    model not in self.model_already_downloaded
                    and not self.maybe_add_deployment(k8s_name=k8s_name, model=model)
                ):
                    # If I am not the first to wake up, I get the new waiter object and continue waiting
                    waiter = self.deployments_to_wait_for.get(model)
                    continue

                console.put.remote(
                    message=RichConsoleSpinnerMessage(
                        id=request_id,
                        label=f"({request_id}) Done waiting for conflicting K8s deployment",
                        state="stop",
                    )
                )
                break

    def signal(self, k8s_name: str, model: str, error: bool = False) -> None:
        if model not in self.deployments_to_wait_for:
            return

        waiter = self.deployments_to_wait_for.pop(model)
        if waiter.k8s_name != k8s_name:
            raise ValueError(
                f"This environment deployment ({k8s_name}) shouldn't have been created because it is conflicting with deployment {waiter.k8s_name}"
            )
        if not error:
            self.model_already_downloaded.add(model)
        waiter.model_downloaded_event.set()



# Copyright (c) IBM Corporation
# SPDX-License-Identifier: MIT

import logging
import os
import subprocess
import time
import uuid
from typing import Any

from ado_actuators.vllm_performance.vllm_performance_test.get_benchmark_results import (
    VLLMBenchmarkResultReadError,
    get_results,
)


class VLLMBenchmarkError(Exception):
    """Raised if there was an issue when running the benchmark"""


def execute_benchmark(
    base_url: str,
    model: str,
    data_set: str,
    interpreter: str = "python",
    num_prompts: int = 500,
    request_rate: int | None = None,
    max_concurrency: int | None = None,
    hf_token: str | None = None,
    benchmark_retries: int = 3,
    retries_timeout: int = 5,
    data_set_path: str | None = None,
    custom_args: dict[str, Any] | None = None,
    burstiness: float = 1,
) -> dict[str, Any]:
    """
    Execute benchmark
    :param base_url: url for vllm endpoint
    :param model: model
    :param data_set: data set name ["sharegpt", "sonnet", "random", "hf"]
    :param interpreter - name of Python interpreter
    :param num_prompts: number of prompts
    :param request_rate: request rate
    :param max_concurrency: max concurrency
    :param hf_token: huggingface token
    :param benchmark_retries: number of benchmark execution retries
    :param retries_timeout: timeout between initial retry
    :param data_set_path: path to the dataset
    :param custom_args: custom arguments to pass to the benchmark.
    keys are vllm benchmark arguments. values are the values to pass to the arguments
    :return: results dictionary

    :raises VLLMBenchmarkError if the benchmark failed to execute after
        benchmark_retries attempts
    """
    logger = logging.getLogger("vllm-bench")

    logger.debug(
        f"executing benchmark, invoking service at {base_url} with the parameters: "
    )
    logger.debug(
        f"model {model}, data set {data_set}, python {interpreter}, num prompts {num_prompts}"
    )
    logger.debug(
        f"request_rate {request_rate}, max_concurrency {max_concurrency}, benchmark retries {benchmark_retries}"
    )
    # The code below is commented as we are switching from a script invocation to command line
    # invocation. If we want to bring back script execution for any reason, this code must be
    # uncommented
    # parameters
    # code = os.path.abspath(
    #    os.path.join(os.path.dirname(__file__), "benchmark_serving.py")
    # )
    # get logger level and forward to subprocess
    log_level = logging.getLevelName(logger.getEffectiveLevel())
    request = f"export HF_TOKEN={hf_token} && " if hf_token is not None else ""
    f_name = f"{uuid.uuid4().hex}.json"
    # Propagate logger's log level to subprocess via env var (if supported)
    request = f"export VLLM_BENCH_LOGLEVEL={log_level} && " + request
    request += (
        # changing from script invocation to cli invocation
        # f"{interpreter} {code} --backend openai --base-url {base_url} --dataset-name {data_set} "
        f"vllm bench serve --backend openai --base-url {base_url} --dataset-name {data_set} "
        f"--model {model} --seed 12345 --num-prompts {num_prompts!s} --save-result --metric-percentiles "
        f'"25,75,99" --percentile-metrics "ttft,tpot,itl,e2el" --result-dir . --result-filename {f_name} '
        f"--burstiness {burstiness} "
    )

    if data_set_path is not None:
        request += f"--dataset-path {data_set_path} "
    if request_rate is not None:
        request += f"--request-rate {request_rate!s} "
    if max_concurrency is not None:
        request += f"--max-concurrency {max_concurrency!s} "
    if custom_args is not None:
        for key, value in custom_args.items():
            request += f"{key} {value!s} "
    timeout = retries_timeout

    logger.debug(f"Command line: {request}")

    for i in range(benchmark_retries):
        try:
            subprocess.check_call(request, shell=True)
            break
        except subprocess.CalledProcessError as e:
            logger.warning(f"Command failed with return code {e.returncode}")
            if i < benchmark_retries - 1:
                logger.warning(
                    f"Will try again after {timeout} seconds. {benchmark_retries - 1 - i} retries remaining"
                )
                time.sleep(timeout)
                timeout *= 2
            else:
                logger.error(
                    f"Failed to execute benchmark after {benchmark_retries} attempts"
                )
                raise VLLMBenchmarkError(f"Failed to execute benchmark {e}")

    try:
        retval = get_results(f_name=f_name)
    except VLLMBenchmarkResultReadError:
        raise VLLMBenchmarkError from VLLMBenchmarkResultReadError

    return retval


def execute_random_benchmark(
    base_url: str,
    model: str,
    num_prompts: int = 500,
    request_rate: int | None = None,
    max_concurrency: int | None = None,
    hf_token: str | None = None,
    benchmark_retries: int = 3,
    retries_timeout: int = 5,
    burstiness: float = 1,
    number_input_tokens: int | None = None,
    max_output_tokens: int | None = None,
    interpreter: str = "python",
) -> dict[str, Any]:
    """
    Execute benchmark with random dataset
    :param base_url: url for vllm endpoint
    :param model: model
    :param data_set: data set name ["sharegpt", "sonnet", "random", "hf"]
    :param hf_token: huggingface token
    :param benchmark_retries: number of benchmark execution retries
    :param retries_timeout: timeout between initial retry
    :param input_token_length: length of input tokens
    :param output_token_length: length of output tokens
    :return: results dictionary
    """
    # Call execute_benchmark with the appropriate arguments
    return execute_benchmark(
        base_url=base_url,
        model=model,
        data_set="random",
        interpreter=interpreter,
        num_prompts=num_prompts,
        request_rate=request_rate,
        max_concurrency=max_concurrency,
        hf_token=hf_token,
        benchmark_retries=benchmark_retries,
        retries_timeout=retries_timeout,
        burstiness=burstiness,
        custom_args={
            "--random-input-len": number_input_tokens,
            "--random-output-len": max_output_tokens,
        },
    )


if __name__ == "__main__":
    results = execute_benchmark(
        interpreter="python3.10",
        base_url="http://localhost:28015",
        data_set="random",
        model="openai/gpt-oss-20b",
        request_rate=None,
        max_concurrency=None,
        hf_token=os.getenv("HF_TOKEN"),
        num_prompts=100,
    )
    print(results)



# Copyright (c) IBM Corporation
# SPDX-License-Identifier: MIT

from typing import Any

import pydantic

from orchestrator.core.actuatorconfiguration.config import GenericActuatorParameters
from orchestrator.modules.operators.base import (
    warn_deprecated_operator_parameters_model_in_use,
)


# In case we need parameters for our actuator, we create a class
# that inherits from GenericActuatorParameters and reference it
# in the parameters_class class variable of our actuator.
# This class inherits from pydantic.BaseModel.
class VLLMPerformanceTestParameters(GenericActuatorParameters):
    namespace: str | None = pydantic.Field(
        default=None,
        description="K8s namespace for running VLLM pod. If not supplied vllm deployments cannot be created.",
    )
    in_cluster: bool = pydantic.Field(
        default=False,
        description="flag to determine whether we are running in K8s cluster or locally",
    )
    verify_ssl: bool = pydantic.Field(
        default=False, description="flag to verify SLL when connecting to server"
    )
    image_secret: str = pydantic.Field(
        default="", description="secret to use when loading image"
    )
    node_selector: dict[str, str] = pydantic.Field(
        default={}, description="dictionary containing node selector key:value pairs"
    )
    deployment_template: str | None = pydantic.Field(
        default=None, description="name of deployment template"
    )
    service_template: str | None = pydantic.Field(
        default=None, description="name of service template"
    )
    pvc_template: str | None = pydantic.Field(
        default=None, description="name of pvc template"
    )
    pvc_name: None | str = pydantic.Field(
        default=None, description="name of pvc to be created/attached"
    )
    interpreter: str = pydantic.Field(
        default="python3", description="name of python interpreter"
    )
    benchmark_retries: int = pydantic.Field(
        default=3, description="number of retries for running benchmark"
    )
    retries_timeout: int = pydantic.Field(
        default=5, description="initial timeout between retries"
    )
    hf_token: str = pydantic.Field(
        default="",
        validate_default=True,
        description="Huggingface token - can be empty if you are accessing fully open models",
    )
    max_environments: int = pydantic.Field(
        default=1, description="Maximum amount of concurrent environments"
    )

    @pydantic.model_validator(mode="before")
    @classmethod
    def make_node_selector_dict(cls, values: Any):
        import json
        from json import JSONDecodeError

        updated = False
        if isinstance(values, dict):
            node_selector = values.get("node_selector", None)
            if node_selector is not None and isinstance(node_selector, str):
                try:
                    values["node_selector"] = (
                        {} if len(node_selector) == 0 else json.loads(node_selector)
                    )
                except JSONDecodeError:
                    raise ValueError(
                        "The node_selector field does not contain a valid dict"
                    )
                updated = True
        elif isinstance(values, GenericActuatorParameters):
            try:
                node_selector = values.node_selector
                if isinstance(node_selector, str):
                    values.node_selector = (
                        {} if len(node_selector) == 0 else json.loads(node_selector)
                    )
                    updated = True
            except JSONDecodeError:
                raise ValueError(
                    "The node_selector field does not contain a valid dict"
                )
            except AttributeError:
                pass
        if updated:
            warn_deprecated_operator_parameters_model_in_use(
                affected_operator="vllm_performance",
                deprecated_from_operator_version="v1.2.2",
                removed_from_operator_version="v1.3",
                latest_format_documentation_url="https://example.com",
            )
        return values



# Copyright (c) IBM Corporation
# SPDX-License-Identifier: MIT


class K8sEnvironmentCreationError(Exception):
    """Error raised when K8 environment cannot be created for some reason"""


class K8sConnectionError(Exception):
    """Error raised when there is an issue connecting to K8s or a service its hosting"""



# Copyright (c) IBM Corporation
# SPDX-License-Identifier: MIT

import logging
import uuid
from enum import Enum
from pathlib import Path
from typing import Any

import yaml

PVC_MOUNT_PATH = "/dev/cache"
PVC_NAME = "vllm-support"
logger = logging.getLogger(__name__)

DEFAULT_PVC_TEMPLATE = "pvc.yaml"
DEFAULT_DEPLOYMENT_TEMPLATE = "deployment.yaml"
DEFAULT_SERVICE_TEMPLATE = "service.yaml"


class VLLMDtype(Enum):
    """
    Type for VLLM
    """

    AUTO = "auto"
    HALF = "half"
    FLOAT16 = "float16"
    BFLOAT16 = "bfloat16"
    FLOAT = "float"
    FLOAT32 = "float32"


class ComponentsYaml:
    """
    Build components Yamls class
    """

    @staticmethod
    def get_k8s_name(model: str) -> str:
        """
        convert model for kubernetes usage
        :param model: LLM model
        :return: K8s unique name for a given LLM model
        """
        m_parts = model.split("/")

        # Making sure the resulting name is not longer than 63 characters as it is
        # the maximum allowed for a name in kubernetes.
        name_prefix = m_parts[-1][: min(len(m_parts[-1]), 25)].rstrip("-")
        return f"vllm-{name_prefix.lower()}-{uuid.uuid4().hex}".replace(".", "-")

    @staticmethod
    def _adjust_file_name(f: str) -> Path:
        """
        Adjust file name to local directory
        :param f: file name
        :return: adjusted file name (Path)
        """
        parent_path = Path(__file__).parents[0]
        return parent_path / f

    def _get_full_path_from_cwd(f: str) -> Path:
        """
        Adjust file name to cwd
        :param f: file name
        :return: adjusted file name (Path)
        """
        return Path.cwd() / f

    @staticmethod
    def deployment_yaml(
        k8s_name: str,
        model: str,
        gpu_type: str = "NVIDIA-A100-80GB-PCIe",
        node_selector: dict[str, str] = {},
        image: str = "vllm/vllm-openai:v0.6.3",
        image_secret: str = "",
        n_gpus: int = 1,
        n_cpus: int = 8,
        memory: str = "128Gi",
        max_batch_tokens: int = 16384,
        gpu_memory_utilization: float = 0.9,
        dtype: VLLMDtype = VLLMDtype.AUTO,
        cpu_offload: int = 0,
        max_num_seq: int = 256,
        template: str | None = None,
        claim_name: str | None = None,
        hf_token: str | None = None,
    ) -> dict[str, Any]:
        """
        Generate deployment yaml
        :param k8s_name: deployment name
        :param model: model name
        :param gpu_type: gpu type, for example NVIDIA-A100-80GB-PCIe, Tesla-V100-PCIE-16GB, etc.
        :param node_selector: optional node selector
        :param image: image name to use
        :param image_secret: name of the image pull secret
        :param n_gpus: number of GPUs to use in VLLM
        :param n_cpus: number of CPUs for VLLM pod
        :param memory: memory for VLLM pod
        :param max_batch_tokens: Vllm parameter - maximum number of batched tokens per iteration
        :param gpu_memory_utilization: VLLM parameter - GPU memory utilization
        :param dtype: VLLM parameter - data type for model weights and activations
        :param cpu_offload: VLLM parameter - the space in GiB to offload to CPU, per GPU
        :param max_num_seq: VLLM parameter - Maximum number of sequences per iteration.
        :param template: template for deployment yaml
        :param claim_name: PVC name
        :param hf_token: huggingface token
        :return:
        """
        # read template
        if template is None:
            logger.debug("Using default Deployment template")
            template_file = ComponentsYaml._adjust_file_name(
                DEFAULT_DEPLOYMENT_TEMPLATE
            )
        else:
            template_file = ComponentsYaml._get_full_path_from_cwd(template)
        logger.debug(
            f"Creating Deployment from template file: {template_file.absolute()}"
        )

        try:
            template_data = template_file.read_text()
            deployment_yaml = yaml.safe_load(template_data)
        except Exception as exception:
            error_string = f"Exception reading deployment yaml template {exception}"
            logger.error(error_string)
            raise ValueError(error_string)

        # Update metadata
        metadata = deployment_yaml["metadata"]
        metadata["name"] = k8s_name
        metadata["labels"]["app.kubernetes.io/instance"] = k8s_name

        # update spec
        spec = deployment_yaml["spec"]
        # selector
        spec["selector"]["matchLabels"]["app.kubernetes.io/instance"] = k8s_name

        # update template
        d_template = spec["template"]
        # template metadata
        d_template["metadata"]["labels"]["app.kubernetes.io/instance"] = k8s_name

        # update template spec
        spec = d_template["spec"]
        # node selector
        spec["nodeSelector"] = {"nvidia.com/gpu.product": gpu_type}

        if len(node_selector) > 0:
            spec["nodeSelector"].update(node_selector)
        # image pull secret
        if image_secret is not None and image_secret != "":
            spec["imagePullSecrets"] = [{"name": image_secret}]
        # volumes
        if claim_name is not None:
            spec["volumes"].extend(
                [{"name": PVC_NAME, "persistentVolumeClaim": {"claimName": claim_name}}]
            )

        # container
        container = spec["containers"][0]
        # image
        container["image"] = image
        # resources
        requests = container["resources"]["requests"]
        requests["cpu"] = str(n_cpus)
        requests["memory"] = memory
        requests["nvidia.com/gpu"] = str(n_gpus)
        limits = container["resources"]["limits"]
        limits["cpu"] = str(n_cpus)
        limits["memory"] = memory
        limits["nvidia.com/gpu"] = str(n_gpus)

        # command
        vllm_serve_args = [
            model,
            "--max-num-batched-tokens",
            f"{max_batch_tokens}",
            "--gpu-memory-utilization",
            f"{gpu_memory_utilization}",
            "--cpu-offload-gb",
            f"{cpu_offload}",
            "--max-num-seq",
            f"{max_num_seq}",
            "--tensor-parallel-size",
            f"{n_gpus}",
            "--dtype",
            dtype.value,
        ]
        container["command"] = ["vllm", "serve"]
        container["args"] = vllm_serve_args

        container["env"] = []
        if hf_token is not None:
            container["env"].extend([{"name": "HF_TOKEN", "value": hf_token}])
        if claim_name is not None:
            container["env"].extend(
                [
                    {
                        "name": "HOME",
                        "value": "/tmp",
                    },
                    {
                        "name": "HF_HOME",
                        "value": f"{PVC_MOUNT_PATH}/transformers_cache",
                    },
                ]
            )
        if logging.root.level == logging.DEBUG:
            container["env"].append({"name": "VLLM_LOGGING_LEVEL", "value": "DEBUG"})
        # volume mounts
        if claim_name is not None:
            container["volumeMounts"].extend(
                [
                    {"name": PVC_NAME, "mountPath": PVC_MOUNT_PATH},
                ]
            )

        # return
        return deployment_yaml

    @staticmethod
    def service_yaml(k8s_name: str, template: str | None = None) -> dict[str, Any]:
        """
        Generate service yaml for a given model
        :param k8s_name: K8s unique name
        :param template: template for service yaml
        :return: service yaml
        """
        # read template
        if template is None:
            logger.debug("Using default Service template")
            template_file = ComponentsYaml._adjust_file_name(DEFAULT_SERVICE_TEMPLATE)
        else:
            template_file = ComponentsYaml._get_full_path_from_cwd(template)
        logger.debug(f"Creating Service from template file: {template_file.absolute()}")

        try:
            template_data = template_file.read_text()
            service_yaml = yaml.safe_load(template_data)
        except Exception as exception:
            error_string = f"Exception reading service yaml template {exception}"
            logger.error(error_string)
            raise ValueError(error_string)

        # Update metadata
        metadata = service_yaml["metadata"]
        metadata["name"] = k8s_name
        metadata["labels"]["app.kubernetes.io/instance"] = k8s_name

        # update selector
        service_yaml["spec"]["selector"]["app.kubernetes.io/instance"] = k8s_name

        # return
        return service_yaml

    @staticmethod
    def pvc_yaml(pvc_name: str, template: str | None = None) -> dict[str, Any]:
        """
        Generate pvc yaml
        :param pvc_name: name of the PVC claim
        :param template: template for pvc yaml
        :return: pvc yaml
        """
        # read template
        if template is None:
            logger.debug("Using default PVC template")
            template_file = ComponentsYaml._adjust_file_name(DEFAULT_PVC_TEMPLATE)
        else:
            template_file = ComponentsYaml._get_full_path_from_cwd(template)
        logger.debug(f"Creating PVC from template file: {template_file.absolute()}")

        try:
            template_data = template_file.read_text()
            pvc_yaml = yaml.safe_load(template_data)
        except Exception as exception:
            error_string = f"Exception reading pvc yaml template {exception}"
            logger.error(error_string)
            raise ValueError(error_string)

        # Update metadata
        pvc_yaml["metadata"]["name"] = pvc_name

        # return
        return pvc_yaml


if __name__ == "__main__":
    t_model = "meta-llama/Llama-3.1-8B-Instruct"
    t_k8s_name = ComponentsYaml.get_k8s_name(model=t_model)
    deployment = ComponentsYaml.deployment_yaml(
        k8s_name=t_k8s_name,
        model=t_model,
        claim_name="vllm-support",
        node_selector={"kubernetes.io/hostname": "cpu16"},
        hf_token="token",
        image="quay.io/dataprep1/data-prep-kit/vllm_image:0.1",
    )
    print(f"Deployment YAML: \n{yaml.dump(deployment)}")
    service = ComponentsYaml.service_yaml(k8s_name=t_k8s_name)
    print(f"Service YAML: \n{yaml.dump(service)}")
    pvc = ComponentsYaml.pvc_yaml(pvc_name="vllm-pvc")
    print(f"PVC YAML: \n{yaml.dump(pvc)}")



# Copyright (c) IBM Corporation
# SPDX-License-Identifier: MIT

import json
import logging
import subprocess
import time

import ray
from ado_actuators.vllm_performance.actuator_parameters import (
    VLLMPerformanceTestParameters,
)
from ado_actuators.vllm_performance.env_manager import (
    Environment,
    EnvironmentManager,
    EnvironmentState,
)
from ado_actuators.vllm_performance.k8s import (
    K8sConnectionError,
    K8sEnvironmentCreationError,
)
from ado_actuators.vllm_performance.k8s.create_environment import (
    create_test_environment,
)
from ado_actuators.vllm_performance.k8s.yaml_support.build_components import (
    VLLMDtype,
)
from ado_actuators.vllm_performance.vllm_performance_test.execute_benchmark import (
    VLLMBenchmarkError,
    execute_random_benchmark,
)
from ray.actor import ActorHandle

from orchestrator.modules.actuators.measurement_queue import MeasurementQueue
from orchestrator.modules.operators.console_output import RichConsoleSpinnerMessage
from orchestrator.schema.experiment import Experiment, ParameterizedExperiment
from orchestrator.schema.request import MeasurementRequest
from orchestrator.utilities.support import (
    compute_measurement_status,
    create_measurement_result,
    dict_to_measurements,
)

logger = logging.getLogger(__name__)


def _build_entity_env(values: dict[str, str]) -> str:
    """
    This is the list of entity parameters that define the environment:
        * model name
        * image name
        * number of gpus
        * gpu type
        * number of cpus
        * memory
        * max batch tokens
        * max number of sequences
        * gpu memory utilization
        * data type
        * cpu offload
    Build entity based environment parameters
    :param values: experiment values
    :return: definition
    """
    env_values = {
        "model": values.get("model"),
        "image": values.get("image"),
        "n_gpus": values.get("n_gpus"),
        "gpu_type": values.get("gpu_type"),
        "n_cpus": values.get("n_cpus"),
        "memory": values.get("memory"),
        "max_batch_tokens": values.get("max_batch_tokens"),
        "gpu_memory_utilization": values.get("gpu_memory_utilization"),
        "dtype": values.get("dtype"),
        "cpu_offload": values.get("cpu_offload"),
        "max_num_seq": values.get("max_num_seq"),
    }
    return json.dumps(env_values)


def _create_environment(
    values: dict[str, str],
    actuator: VLLMPerformanceTestParameters,
    node_selector: dict[str, str],
    request_id: str,
    env_manager: ActorHandle[EnvironmentManager],
    check_interval: int = 5,
    timeout: int = 1200,
) -> tuple[str, str]:
    """
     Create environment

     Important: This function will block until env_manager.get_environment
     returns an environment.
     The env_manager will not return an environment until there is one free
     to be used

     :param values: experiment values
     :param actuator: actuator parameters
     :param node_selector: node selector
     :param request_id the request associated with this environment
     :param env_manager: environment manager
     :param check_interval: wait interval
     :param timeout: timeout
    :return: kubernetes environment name

    :raises K8sEnvironmentCreationError if there was an issue
    - If the creation step fails after three attempts
    - If after creation the environment was not in ready state after timeout seconds (1200 default)

    """
    from orchestrator.modules.operators.console_output import (
        RichConsoleSpinnerMessage,
    )

    console = ray.get_actor(name="RichConsoleQueue")
    environment_usage = ray.get(env_manager.environment_usage.remote())

    # get model for experiment
    model = values.get("model")

    # create environment definition
    definition = _build_entity_env(values=values)
    console.put.remote(
        message=RichConsoleSpinnerMessage(
            id=request_id,
            label=f"({request_id}) Waiting for deployment environment slot to be available - total slots {environment_usage.get('max')}",
            state="start",
        )
    )
    while True:

        try:
            env: Environment = ray.get(
                env_manager.get_environment.remote(model=model, definition=definition)
            )
        except Exception as e:
            raise e
        if env is not None:
            console.put.remote(
                message=RichConsoleSpinnerMessage(
                    id=request_id,
                    label=f"{request_id} Got environment slot {env.k8s_name}",
                    state="stop",
                )
            )
            break

        # This is to guarantee that the request is next in line as soon as an environment is available
        ray.get(env_manager.wait_for_env.remote())

    error = None
    logger.debug(
        f"Environment state {env.state}, name {env.k8s_name}, definition {definition}"
    )

    start = time.time()

    # We retrieve the PVC name from the actor because it is one to be shared for the whole experiment
    pvc_name = ray.get(env_manager.get_experiment_pvc_name.remote())

    match env.state:
        case EnvironmentState.NONE:

            # Environment does not exist, create it
            logger.debug(f"Environment {env.k8s_name} does not exist. Creating it")
            tmout = 1

            # To avoid data corruption we wait if another environment is concurrently downloading the same model for the first time
            ray.get(
                env_manager.wait_deployment_before_starting.remote(
                    env=env, request_id=request_id
                )
            )

            for attempt in range(3):
                console.put.remote(
                    message=RichConsoleSpinnerMessage(
                        id=request_id,
                        label=f"({request_id}) Creating vLLM deployment {env.k8s_name} (attempt {attempt+1}/3)...",
                        state="start",
                    )
                )
                try:
                    create_test_environment(
                        k8s_name=env.k8s_name,
                        model=model,
                        in_cluster=actuator.in_cluster,
                        verify_ssl=actuator.verify_ssl,
                        image=values.get("image"),
                        image_secret=actuator.image_secret,
                        deployment_template=actuator.deployment_template,
                        service_template=actuator.service_template,
                        n_gpus=int(values.get("n_gpus")),
                        gpu_type=values.get("gpu_type"),
                        node_selector=node_selector,
                        n_cpus=int(values.get("n_cpus")),
                        memory=values.get("memory"),
                        max_batch_tokens=int(values.get("max_batch_tokens")),
                        gpu_memory_utilization=float(
                            values.get("gpu_memory_utilization")
                        ),
                        dtype=VLLMDtype(values.get("dtype", "auto")),
                        cpu_offload=int(values.get("cpu_offload")),
                        max_num_seq=int(values.get("max_num_seq")),
                        hf_token=actuator.hf_token,
                        reuse_service=False,
                        reuse_deployment=False,
                        namespace=actuator.namespace,
                        pvc_name=pvc_name,
                        check_interval=check_interval,
                        timeout=timeout,
                    )
                    # Update manager
                    env_manager.done_creating.remote(identifier=env.k8s_name)
                    error = None
                    break
                except Exception as e:
                    logger.error(
                        f"Attempt {attempt}. Failed to create test environment {e}"
                    )
                    error = f"Failed to create test environment {e}"
                    time.sleep(tmout)
                    tmout *= 2

            # Check if error after three attempts
            if error is None:
                console.put.remote(
                    message=RichConsoleSpinnerMessage(
                        id=request_id,
                        label=f"({request_id})  Created vLLM deployment {env.k8s_name}",
                        state="stop",
                    )
                )
                logger.info(
                    f"Created test environment {env.k8s_name} in {time.time() - start} sec"
                )
            else:
                console.put.remote(
                    message=RichConsoleSpinnerMessage(
                        id=request_id,
                        label=f"({request_id}) Failed to create {env.k8s_name}. Aborting.",
                        state="stop",
                    )
                )

                # In case of failure creating the environment deployment we must release any
                # other request with a deployment conflicting with this request's deployment
                # We also need to release the slot for this environment
                ray.get(
                    env_manager.cleanup_failed_deployment.remote(
                        identifier=env.k8s_name
                    )
                )

                raise K8sEnvironmentCreationError(
                    f"Failed to create test environment {env.k8s_name}: {error}"
                )

    return env.k8s_name, definition


def _connect_to_vllm_server(
    k8s_name: str,
    actuator_parameters: VLLMPerformanceTestParameters,
    port: int,
) -> tuple[str, subprocess.Popen | None]:
    """Returns the URL of the vLLM inference server

    Creates a port forward for the inference server if test
    is not running on the cluster with the service

    Parameters:
        k8s_name: The name of the vLLM service
        actuator_parameters: VLLMPerformanceTestParameters instance containing
            namespace and test location (in_cluster or not) information

    Returns:
        A tuple containing
        - The URL of the created vLLM server
        - If a port-forward is created the POpen object for the port-forward
          Otherwise None

    Raise:
        K8ConnectionError if a port-forward could not be created
    """

    # create environment
    if not actuator_parameters.in_cluster:
        logger.info("We are running locally connecting to remote cluster")
        logger.info("please make sure that you have executed `oc login`")
        logger.info(
            "We are using ports from 10000 and above to communicate with the cluster, "
            "please make sure that it is not in use"
        )

    if actuator_parameters.in_cluster:
        # we are running in cluster, connect to service directly
        base_url = (
            f"http://{k8s_name}.{actuator_parameters.namespace}.svc.cluster.local:80"
        )
        pf = None
    else:
        # we are running locally. need to do port-forward and connect to the local one
        pf_command_args = [
            "kubectl",
            "port-forward",
            f"svc/{k8s_name}",
            "-n",
            f"{actuator_parameters.namespace}",
            f"{port}:80",
        ]
        try:
            pf = subprocess.Popen(
                pf_command_args,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
            # make sure that port forwarding is up
            time.sleep(5)
            # Check if there is a returncode- if there is it means port-forward exited
            if pf.returncode:
                raise K8sConnectionError(
                    f"failed to start port forward to service {k8s_name} - port-forward command exited for unknown reason. Check logs."
                )
        except Exception as e:
            logger.warning(f"failed to start port forward to service {k8s_name} - {e}")
            raise K8sConnectionError(
                f"failed to start port forward to service {k8s_name} - {e}"
            )

        base_url = f"http://localhost:{port}"

    return base_url, pf


@ray.remote
def run_resource_and_workload_experiment(
    request: MeasurementRequest,
    experiment: Experiment | ParameterizedExperiment,
    state_update_queue: MeasurementQueue,
    actuator_parameters: VLLMPerformanceTestParameters,
    node_selector: dict[str, str],
    env_manager: ActorHandle,
    local_port: int,
):
    """
    Runs an experiment on a specific compute resource and inference workload configuration.

    This requires spinning up a vLLM instance with the given compute resources

    :param request: measurement request
    :param experiment: definition of experiment
    :param state_update_queue: update queue
    :param actuator_parameters: actuator parameters
    :param node_selector: node selector
    :param env_manager: environment manager
    :param local_port: local port to use
    :return:
    """

    # This function
    # 1. Performs the measurement represented by MeasurementRequest
    # 2. Updates MeasurementRequest with the results of the measurement and status
    # 3. Puts it in the stateUpdateQueue

    logger.debug(
        f"number of entities {len(request.entities)}, actuator parameters {actuator_parameters}, node selector {node_selector}"
    )

    # placeholder for measurements
    measurements = []
    current_port = local_port - 1
    console = ray.get_actor(name="RichConsoleQueue")

    # For every entity
    for entity in request.entities:

        port_forward = None
        definition = None
        started_benchmarking = False
        try:
            values = experiment.propertyValuesFromEntity(entity=entity)

            logger.info(f"Creating K8s environment for {entity.identifier}")

            # Will raise an K8sEnvironmentCreationError if the environment could not be created
            k8s_name, definition = _create_environment(
                values=values,
                actuator=actuator_parameters,
                node_selector=node_selector,
                env_manager=env_manager,
                request_id=request.requestid,
            )

            # Will raise an K8sConnectionError if a port-forward was required
            # but could not be created
            current_port += 1
            base_url, port_forward = _connect_to_vllm_server(
                k8s_name, actuator_parameters, current_port
            )

            logger.info(f"Will use vllm server at {base_url}")

            request_rate = int(values.get("request_rate"))
            if request_rate < 0:
                request_rate = None
            max_concurrency = int(values.get("max_concurrency"))
            if max_concurrency < 0:
                max_concurrency = None

            started_benchmarking = True
            console.put.remote(
                message=RichConsoleSpinnerMessage(
                    id=request.requestid,
                    label=f"({request.requestid}) Executing vllm bench serve",
                    state="start",
                )
            )
            result = execute_random_benchmark(
                base_url=base_url,
                model=values.get("model"),
                interpreter=actuator_parameters.interpreter,
                num_prompts=int(values.get("num_prompts")),
                request_rate=request_rate,
                max_concurrency=max_concurrency,
                hf_token=actuator_parameters.hf_token,
                benchmark_retries=actuator_parameters.benchmark_retries,
                retries_timeout=actuator_parameters.retries_timeout,
                number_input_tokens=int(values.get("number_input_tokens")),
                max_output_tokens=int(values.get("max_output_tokens")),
                burstiness=float(values.get("burstiness")),
            )

        except (
            K8sEnvironmentCreationError,
            K8sConnectionError,
            VLLMBenchmarkError,
        ) as error:
            logger.error(f"Error running tests for entity {entity.identifier}: {error}")
            measurements.append(
                create_measurement_result(
                    identifier=entity.identifier,
                    measurements=[],
                    error=str(error),
                    reference=request.experimentReference,
                )
            )
        except Exception as error:
            logger.critical(f"Unexpected error for entity {entity.identifier}: {error}")
            measurements.append(
                create_measurement_result(
                    identifier=entity.identifier,
                    measurements=[],
                    error=f"Unexpected error in experiment execution: {error}",
                    reference=request.experimentReference,
                )
            )
        else:
            measured_values = dict_to_measurements(
                results=result, experiment=experiment
            )
            measurements.append(
                create_measurement_result(
                    identifier=entity.identifier,
                    measurements=measured_values,
                    error=None,
                    reference=request.experimentReference,
                )
            )
        finally:
            if started_benchmarking:
                console.put.remote(
                    message=RichConsoleSpinnerMessage(
                        id=request.requestid,
                        label=f"({request.requestid}) Completed benchmark",
                        state="stop",
                    )
                )
            if port_forward is not None:
                port_forward.kill()
            if definition is not None:
                env_manager.done_using.remote(identifier=k8s_name)

    # For multi entity experiments if ONE entity had ValidResults the status must be SUCCESS
    if len(measurements) > 0:
        request.measurements = measurements
    request.status = compute_measurement_status(measurements=measurements)
    logger.debug(f"request status is {request.status}. pushing to update queue")
    # Push the request to the state updates queue
    state_update_queue.put(request, block=False)


@ray.remote
def run_workload_experiment(
    request: MeasurementRequest,
    experiment: Experiment | ParameterizedExperiment,
    state_update_queue: MeasurementQueue,
    actuator_parameters: VLLMPerformanceTestParameters,
):
    """
    Runs an experiment with a specific inference workload configuration on a given endpoint.

    The compute resource associated with the end-point is not known.

    :param request: measurement request
    :param experiment: definition of experiment
    :param state_update_queue: update queue
    :param actuator_parameters: actuator parameters
    :return:
    """

    # This function
    # 1. Performs the measurement represented by MeasurementRequest
    # 2. Updates MeasurementRequest with the results of the measurement and status
    # 3. Puts it in the stateUpdateQueue

    # placeholder for measurements
    measurements = []
    # For every entity
    for entity in request.entities:
        measured_values = []
        error = None
        try:
            values = experiment.propertyValuesFromEntity(entity=entity)
            logger.debug(
                f"Values for entity {entity.identifier} and experiment {experiment.identifier} "
                f"experiment type is {type(experiment)} are {json.dumps(values)}"
            )

            request_rate = int(values.get("request_rate"))
            if request_rate < 0:
                request_rate = None
            max_concurrency = int(values.get("max_concurrency"))
            if max_concurrency < 0:
                max_concurrency = None

            # Will raise VLLMBenchmarkError if there is a problem
            result = execute_random_benchmark(
                base_url=values.get("endpoint"),
                model=values.get("model"),
                interpreter=actuator_parameters.interpreter,
                num_prompts=int(values.get("num_prompts")),
                request_rate=request_rate,
                max_concurrency=max_concurrency,
                hf_token=actuator_parameters.hf_token,
                benchmark_retries=actuator_parameters.benchmark_retries,
                retries_timeout=actuator_parameters.retries_timeout,
                number_input_tokens=int(values.get("number_input_tokens")),
                max_output_tokens=int(values.get("max_output_tokens")),
                burstiness=float(values.get("burstiness")),
            )
        except VLLMBenchmarkError as e:
            error = f"Encountered benchmark error when testing entity {entity.identifier}: {e}"
            logger.error(error)
        except Exception as e:
            error = f"Unexpected error for entity {entity.identifier}: {e}"
            logger.error(error)
        else:
            measured_values = dict_to_measurements(
                results=result, experiment=experiment
            )
            logger.debug(f"measured values {measured_values}")
        finally:
            measurements.append(
                create_measurement_result(
                    identifier=entity.identifier,
                    measurements=measured_values,
                    error=error,
                    reference=request.experimentReference,
                )
            )

    # For multi entity experiments if ONE entity had ValidResults the status must be SUCCESS
    if len(measurements) > 0:
        request.measurements = measurements
    request.status = compute_measurement_status(measurements=measurements)
    logger.debug(f"request status is {request.status}. pushing to update queue")
    # Push the request to the state updates queue
    state_update_queue.put(request, block=False)



# Copyright (c) IBM Corporation
# SPDX-License-Identifier: MIT

import asyncio
import logging
from enum import Enum

import ray
from ado_actuators.vllm_performance.deployment_management import (
    DeploymentConflictManager,
)
from ado_actuators.vllm_performance.k8s import K8sEnvironmentCreationError
from ado_actuators.vllm_performance.k8s.manage_components import (
    ComponentsManager,
)
from ado_actuators.vllm_performance.k8s.yaml_support.build_components import (
    ComponentsYaml,
)
from kubernetes.client import ApiException

logger = logging.getLogger(__name__)


class EnvironmentState(Enum):
    """
    Environment state
    """

    NONE = "None"
    READY = "ready"


class Environment:
    """
    environment class
    """

    def __init__(self, model: str, configuration: str):
        """
        Defines an environment for a model
        :param model: LLM model name
        :param configuration: The full deployment configuration
        """
        self.k8s_name = ComponentsYaml.get_k8s_name(model=model)
        self.state = EnvironmentState.NONE
        self.configuration = configuration
        self.model = model


class EnvironmentsQueue:
    def __init__(self):
        self.environments_queue = []

    async def wait(self) -> None:
        wait_event = asyncio.Event()
        self.environments_queue.append(wait_event)
        await wait_event.wait()

    def signal_next(self) -> None:
        if len(self.environments_queue) > 0:
            event = self.environments_queue.pop(0)
            event.set()


@ray.remote
class EnvironmentManager:
    """
    This is a Ray actor (singleton) managing environments
    """

    def __init__(
        self,
        namespace: str,
        max_concurrent: int,
        in_cluster: bool = True,
        verify_ssl: bool = False,
        pvc_name: str | None = None,
        pvc_template: str | None = None,
    ):
        """
        Initialize
        :param namespace: deployment namespace
        :param max_concurrent: maximum amount of concurrent environment
        :param in_cluster: flag in cluster
        :param verify_ssl: flag verify SSL
        :param pvc_name: name of the PVC to be created / used
        :param pvc_template: template of the PVC to be created
        """
        self.in_use_environments: dict[str, Environment] = {}
        self.free_environments: list[Environment] = []
        self.environments_queue = EnvironmentsQueue()
        self.deployment_conflict_manager = DeploymentConflictManager()
        self.namespace = namespace
        self.max_concurrent = max_concurrent
        self.in_cluster = in_cluster
        self.verify_ssl = verify_ssl

        # component manager for cleanup
        self.manager = ComponentsManager(
            namespace=self.namespace,
            in_cluster=self.in_cluster,
            verify_ssl=self.verify_ssl,
            init_pvc=True,
            pvc_name=pvc_name,
            pvc_template=pvc_template,
        )

    def _delete_environment_k8s_resources(self, k8s_name: str) -> None:
        """
        Deletes a deployment. Intended to be used for cleanup or error recovery
        param: identifier: the deployment identifier
        """
        try:
            self.manager.delete_service(k8s_name=k8s_name)
        except ApiException as e:
            if e.reason != "Not Found":
                raise e
        try:
            self.manager.delete_deployment(k8s_name=k8s_name)
        except ApiException as e:
            if e.reason != "Not Found":
                raise e

    def environment_usage(self) -> dict:

        return {"max": self.max_concurrent, "in_use": self.active_environments}

    async def wait_for_env(self):
        await self.environments_queue.wait()

    def get_environment(self, model: str, definition: str) -> Environment | None:
        """
        Get an environment for definition
        :param model: LLM model name
        :param definition: environment definition - json string containing:
                        model, image, n_gpus, gpu_type, n_cpus, memory, max_batch_tokens,
                        gpu_memory_utilization, dtype, cpu_offload, max_num_seq
        :param increment_usage: increment usage flag
        :return: environment state
        """

        # check if there's an existing free environment satisfying the request
        env = self.get_matching_free_environment(definition)
        if env is None:
            if self.active_environments >= self.max_concurrent:
                # can't create more environments now, need clean up
                if len(self.free_environments) == 0:
                    # No room for creating a new environment
                    logger.debug(
                        f"There are already {self.max_concurrent} actively in use, and I can't create a new one"
                    )
                    return None

                # There are unused environments, let's try to evict one
                environment_evicted = False
                eviction_index = 0
                # Continue looping until we find one environment that can be successfully evicted or we have gone through them all
                while not environment_evicted and eviction_index < len(
                    self.free_environments
                ):
                    environment_to_evict = self.free_environments[eviction_index]
                    try:
                        # _delete_environment_k8s_resources will not raise an error if for whatever the reason the service
                        # or the deployment we are trying to delete does not exist anymore, and we assume
                        # the deployment was properly deleted.
                        self._delete_environment_k8s_resources(
                            k8s_name=environment_to_evict.k8s_name
                        )
                    except ApiException as e:
                        # If we can't delete this environment we try with the next one, but we do not
                        # delete the current env from the free list. This is to avoid spawning more pods than the maximum configured
                        # in the case the failing ones are still running.
                        # Since the current eviction candidate environment will stay in the free ones, some other measurement might
                        # try to evict again and perhaps succeed (e.g., connection restored to the cluster).
                        logger.critical(
                            f"Error deleting deployment or service {environment_to_evict.k8s_name}: {e}"
                        )
                        eviction_index += 1
                        continue

                    logger.info(
                        f"deleted environment {environment_to_evict.k8s_name}. "
                        f"Active environments {self.active_environments}"
                    )
                    environment_evicted = True

                if environment_evicted:
                    # successfully deleted an environment
                    self.free_environments.pop(eviction_index)
                elif len(self.in_use_environments) > 0:
                    # all the free ones have failed deleting but there is one or more in use that
                    # might make room for waiting measurements. In this case we just behave as if there
                    # are no free available environments and we wait.
                    return None
                else:
                    # None of the free environments could be evicted due to errors and none are in use
                    # To avoid a deadlock of the operation we fail the measurement
                    raise K8sEnvironmentCreationError(
                        "All free environments failed deleting and none are currently in use."
                    )

            # We either made space or we had enough space already
            env = Environment(model=model, configuration=definition)
            logger.debug(f"New environment created for definition {definition}")

        # If deployments target the same model and the model is not in the HF cache, they would all try to download it.
        # This can lead to corruption of the HF cache data (shared PVC).
        # To avoid this situation, we keep track of the models downloaded by the actuator during the current operation.
        # If a deployment wants to download a model for the first time, we do not allow other deployment using the
        # same model to start in parallel.
        # Once the very first download of a model is done we let any number of deployments using the same model to start
        # in parallel as they would only read the model from the cache.
        self.deployment_conflict_manager.maybe_add_deployment(
            k8s_name=env.k8s_name, model=model
        )

        self.in_use_environments[env.k8s_name] = env

        return env

    @property
    def active_environments(self) -> int:
        return len(self.in_use_environments) + len(self.free_environments)

    def get_experiment_pvc_name(self) -> str:
        return self.manager.pvc_name

    def done_creating(self, identifier: str) -> None:
        """
        Report creation
        :param identifier: environment identifier
        :return: None
        """
        self.in_use_environments[identifier].state = EnvironmentState.READY
        model = self.in_use_environments[identifier].model

        self.deployment_conflict_manager.signal(k8s_name=identifier, model=model)

    def cleanup_failed_deployment(self, identifier: str) -> None:
        env = self.in_use_environments[identifier]
        self._delete_environment_k8s_resources(k8s_name=identifier)
        self.done_using(identifier=identifier, reclaim_on_completion=True)
        self.deployment_conflict_manager.signal(
            k8s_name=identifier, model=env.model, error=True
        )

    def get_matching_free_environment(self, configuration: str) -> Environment | None:
        """
        Find a deployment matching a deployment configuration
        :param configuration: The deployment configuration to match
        :return: An already existing deployment or None
        """
        for id, env in enumerate(self.free_environments):
            if env.configuration == configuration:
                del self.free_environments[id]
                return env
        return None

    async def wait_deployment_before_starting(
        self, env: Environment, request_id: str
    ) -> None:
        await self.deployment_conflict_manager.wait(
            request_id=request_id, k8s_name=env.k8s_name, model=env.model
        )

    def done_using(self, identifier: str, reclaim_on_completion: bool = False) -> None:
        """
        Report test completion
        :param definition: environment definition
        :param reclaim_on_completion: flag to indicate the environment is to be completely removed and not freed for later use
        :return: None
        """
        env = self.in_use_environments.pop(identifier)
        if not reclaim_on_completion:
            self.free_environments.append(env)

        # Wake up any other deployment waiting in the queue for a
        # free environment.
        self.environments_queue.signal_next()

    def cleanup(self) -> None:
        """
        Clean up environment
        :return: None
        """
        logger.info("Cleaning environments")
        all_envs = list(self.in_use_environments.values()) + self.free_environments
        for env in all_envs:
            self._delete_environment_k8s_resources(k8s_name=env.k8s_name)

        # We only delete the PVC if it was created by this actuator
        if self.manager.pvc_created:
            logger.debug("Deleting PVC")
            self.manager.delete_pvc()
        else:
            logger.debug("No PVC was created. Nothing to delete!")



# Copyright (c) IBM Corporation
# SPDX-License-Identifier: MIT

import logging
import os
import uuid

import ray
import yaml
from ado_actuators.vllm_performance.actuator_parameters import (
    VLLMPerformanceTestParameters,
)
from ado_actuators.vllm_performance.env_manager import (
    EnvironmentManager,
)
from ado_actuators.vllm_performance.experiment_executor import (
    run_resource_and_workload_experiment,
    run_workload_experiment,
)

from orchestrator.core.actuatorconfiguration.config import GenericActuatorParameters
from orchestrator.modules.actuators.base import (
    ActuatorBase,
    DeprecatedExperimentError,
    MissingConfigurationForExperimentError,
)
from orchestrator.modules.actuators.catalog import ExperimentCatalog
from orchestrator.modules.actuators.measurement_queue import MeasurementQueue
from orchestrator.modules.operators.orchestrate import CLEANER_ACTOR
from orchestrator.schema.entity import Entity
from orchestrator.schema.experiment import Experiment, ParameterizedExperiment
from orchestrator.schema.reference import ExperimentReference
from orchestrator.schema.request import MeasurementRequest

logger = logging.getLogger(__name__)


# An Actuator must do three things
# 1. Provide a catalog of Experiments it can execute - the catalog method
# 2. Provide a way to run those experiments asynchronously - the submit method
# 3. Provide the results of the experiments - via the Results Queue
@ray.remote
class VLLMPerformanceTest(ActuatorBase):

    identifier = "vllm_performance"  # The user-facing label you want this actuator to be called by
    parameters_class = (
        VLLMPerformanceTestParameters  # we tell ado what our parameters class is
    )

    @classmethod
    def catalog(
        cls, actuator_configuration: GenericActuatorParameters | None = None
    ) -> ExperimentCatalog:
        """Returns the Experiments your actuator provides"""

        # The catalog be formed in code here or read from a file containing the Experiments models
        # This shows reading from a file

        path = os.path.abspath(__file__)
        path = os.path.split(path)[0]
        with open(os.path.join(path, "experiments.yaml")) as f:
            data = yaml.safe_load(f)
            experiments = [Experiment(**data[e]) for e in data]

        return ExperimentCatalog(
            catalogIdentifier=cls.identifier,
            experiments={e: e for e in experiments},
        )

    def __init__(
        self,
        queue: MeasurementQueue,
        params: VLLMPerformanceTestParameters,
    ):
        """
        queue: Queue where experiment results are put for consumers
        params: This actuators configuration parameters (Note: Soon this will be a model defined by the Actuator)
        """

        # Set logging and the ivar self._stateUpdateQueue and self._parameters
        super().__init__(queue=queue, params=params)
        # Set parameters
        self.actuator_parameters = params
        # çreate environment manager actor
        self.env_manager = None

        if self.actuator_parameters.namespace:
            try:
                self.env_manager = EnvironmentManager.remote(
                    namespace=params.namespace,
                    max_concurrent=params.max_environments,
                    in_cluster=params.in_cluster,
                    verify_ssl=params.verify_ssl,
                    pvc_name=params.pvc_name,
                    pvc_template=params.pvc_template,
                )
            except Exception as error:
                self.log.warning(
                    f"Unable to create kubernetes environment manager due to {error}. "
                    f"Will not be able to execute experiments requiring deploying on K8s"
                )
            else:
                # add to clean up
                try:
                    cleaner_handle = ray.get_actor(name=CLEANER_ACTOR)
                    cleaner_handle.add_to_cleanup.remote(handle=self.env_manager)
                except Exception as e:
                    logger.warning(
                        f"Failed to register custom actors for clean up {e}. Make sure you clean it up"
                    )
        else:
            self.log.warning(
                "No namespace set in acutator configuration - will not be able to create deployments"
            )

        # initialize local port
        self.local_port = 10000

    async def submit(
        self,
        entities: list[Entity],
        experimentReference: ExperimentReference,
        requesterid: str,
        requestIndex: int,
    ) -> list[str]:
        """Runs experimentReference on entities

        The result of running the experiment is expected to be returned asynchronously via the stateUpdateQueue
        as a MeasurementRequest instance

        :param entities: A list of Entity instances representing the entities to be measured
        :param experimentReference: An ExperimentReference defining the experiment to run on the entities.
               Expected to come from the actuators catalog.
        :param requesterid: Unique identifier of the requester of this measurement
        :param requestIndex: The index of this request i.e. this is the ith request by the requester in the current run

        Returns:
            A list with the ids of the experiments it submitted.
            NOTE: An actuator may not submit all entities in one batch.

        Raises:
            DeprecatedExperimentError: if the experimentReference points to an experiment that is deprecated.
            MissingConfigurationForExperimentError
        """

        # The following steps MUST be done
        # How they are implemented is up to the developer

        # 1. Create MeasurementRequest(s) instances representing the experiments to be performed
        # 2. Asynchronously launch the experiments -> placing their results in the resultsQueue
        # 3. Return the ids of the MeasurementRequests submit to the caller of this function

        # Here we give an example of creating the MeasurementRequests in this method
        # and executing asynchronously via a remote ray function

        # Create MeasurementRequest(s)
        # You can create one for all entities or one for each entity or any split
        # NOTE: ALL must have same requestIndex - this is used to maintain timeseries of requests
        request = MeasurementRequest(
            operation_id=requesterid,
            requestIndex=requestIndex,
            experimentReference=experimentReference,
            entities=entities,
            requestid=str(uuid.uuid4())[:6],  # Create UID as you like
        )

        # Check if the requested experiment is deprecated
        experiment = self.__class__.catalog().experimentForReference(
            request.experimentReference
        )
        """
        Check if the experiment has parameterization fields and create the right ParameterizedExperiment instance
        Note: this is necessary for getting values of optional properties that the discoverySpace overrides
        """
        if experimentReference.parameterization:
            experiment = ParameterizedExperiment(
                parameterization=experimentReference.parameterization,
                **experiment.model_dump(),
            )

        if experiment.deprecated is True:
            raise DeprecatedExperimentError(f"Experiment {experiment} is deprecated")

        if experiment.identifier == "test-deployment-v1":
            if not self.env_manager:
                raise MissingConfigurationForExperimentError(
                    f"Actuator configuration did not contain sufficient information for a kubernetes environment manager to be created. "
                    f"Experiment {experiment} requires a kubernetes environment manager to be executable."
                )

            # Execute experiment
            # Note: Here the experiment instance is just past for convenience since we retrieved it above
            run_resource_and_workload_experiment.remote(
                request=request,
                experiment=experiment,
                state_update_queue=self._stateUpdateQueue,
                actuator_parameters=self.actuator_parameters,
                node_selector=self.actuator_parameters.node_selector,
                env_manager=self.env_manager,
                local_port=self.local_port,
            )
            self.local_port += len(request.entities)
        else:
            run_workload_experiment.remote(
                request=request,
                experiment=experiment,
                state_update_queue=self._stateUpdateQueue,
                actuator_parameters=self.actuator_parameters,
            )

        return [request.requestid]
